# This one needs run_dir, out_dir, basemask and sample_sheet

from cg.constants.demultiplexing import BclConverter

DEMULTIPLEX_COMMAND = {
    BclConverter.BCL2FASTQ: """
log "singularity exec --bind \
/home/proj/{environment}/sequencing_data/illumina_novaseq_flow_cells/demultiplexed-runs,\
/home/proj/{environment}/sequencing_data/illumina_novaseq_flow_cells/flow_cells,\
/home/proj/{environment}/sequencing_data/illumina_novaseq_flow_cells/flow_cells/'$SLURM_JOB_ID':/run/user/$(id -u) \
/home/proj/{environment}/demux-on-hasta/novaseq/container/bcl2fastq_v2-20-0.sif \
bcl2fastq --loading-threads 3 --processing-threads 15 --writing-threads 3 \
--runfolder-dir {run_dir} --output-dir {unaligned_dir} \
--sample-sheet {sample_sheet} \
--barcode-mismatches 1
touch {demux_completed_file}"

singularity exec --bind \
/home/proj/{environment}/sequencing_data/illumina_novaseq_flow_cells/demultiplexed-runs,\
/home/proj/{environment}/sequencing_data/illumina_novaseq_flow_cells/flow_cells,\
/home/proj/{environment}/sequencing_data/illumina_novaseq_flow_cells/flow_cells/'$SLURM_JOB_ID':/run/user/$(id -u) \
/home/proj/{environment}/demux-on-hasta/novaseq/container/bcl2fastq_v2-20-0.sif \
bcl2fastq --loading-threads 3 --processing-threads 15 --writing-threads 3 \
--runfolder-dir {run_dir} --output-dir {unaligned_dir} \
--sample-sheet {sample_sheet} \
--barcode-mismatches 1
touch {demux_completed_file}
log "bcl2fastq finished!"
""",
    BclConverter.BCLCONVERT: """
log "dragen --bcl-conversion-only true \
--bcl-input-directory {run_dir} \
--output-directory {demux_dir} \
--force
touch {demux_completed_file}"

dragen --bcl-conversion-only true \
--bcl-input-directory {run_dir} \
--output-directory {demux_dir} \
--force
touch {demux_completed_file}
log "Dragen BCL Convert finished!"
""",
}
# This needs flow cell id, email. logfile
DEMULTIPLEX_ERROR = """
mail -s 'ERROR demultiplexing of {flow_cell_id}' {email} < '{logfile}'
if [[ -e {demux_dir} ]]
then
    rm -r {demux_dir}
fi
if [[ -e {demux_started} ]]
then
    rm -r {demux_started}
fi
"""



from typing import Annotated
from pydantic import BaseModel, BeforeValidator


def empty_str_to_none(v: str) -> str | None:
    return v or None


class PicardMarkduplicate(BaseModel):
    insert_size: Annotated[int | None, BeforeValidator(empty_str_to_none)]
    duplication_rate: Annotated[float | None, BeforeValidator(empty_str_to_none)]


class MicrosaltSamtoolsStats(BaseModel):
    total_reads: Annotated[int | None, BeforeValidator(empty_str_to_none)]
    mapped_rate: Annotated[float | None, BeforeValidator(empty_str_to_none)]
    average_coverage: Annotated[float | None, BeforeValidator(empty_str_to_none)]
    coverage_10x: Annotated[float | None, BeforeValidator(empty_str_to_none)]


class SampleMetrics(BaseModel):
    picard_markduplicate: PicardMarkduplicate
    microsalt_samtools_stats: MicrosaltSamtoolsStats


class QualityMetrics(BaseModel):
    samples: dict[str, SampleMetrics]



"""Tests for functions of DemultiplexAPI."""

from pathlib import Path

import pytest

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.meta.demultiplex.housekeeper_storage_functions import (
    add_sample_sheet_path_to_housekeeper,
)
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData


def test_is_sample_sheet_in_housekeeper_exists(
    demultiplexing_context_for_demux: CGConfig, tmp_bcl2fastq_flow_cell: FlowCellDirectoryData
):
    """Test that checking the existence of an existing sample sheet in Housekeeper returns True."""
    # GIVEN a DemultiplexAPI and a flow cell with a sample sheet
    demux_api: DemultiplexingAPI = demultiplexing_context_for_demux.demultiplex_api
    demultiplexing_context_for_demux.illumina_flow_cells_directory = (
        tmp_bcl2fastq_flow_cell.path.parent
    )

    # GIVEN that the sample sheet is in Housekeeper
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=tmp_bcl2fastq_flow_cell.path,
        flow_cell_name=tmp_bcl2fastq_flow_cell.id,
        hk_api=demultiplexing_context_for_demux.housekeeper_api,
    )

    # WHEN testing if the sample sheet is in Housekeeper
    result: bool = demux_api.is_sample_sheet_in_housekeeper(flow_cell_id=tmp_bcl2fastq_flow_cell.id)

    # THEN the sample sheet should be in Housekeeper
    assert result


def test_is_sample_sheet_in_housekeeper_not_in_hk(
    demultiplexing_context_for_demux: CGConfig, tmp_bcl2fastq_flow_cell: FlowCellDirectoryData
):
    """Test that checking the existence of a non-existing sample sheet in Housekeeper returns False."""
    # GIVEN a DemultiplexAPI and a flow cell with a sample sheet
    demux_api: DemultiplexingAPI = demultiplexing_context_for_demux.demultiplex_api
    demultiplexing_context_for_demux.illumina_flow_cells_directory = (
        tmp_bcl2fastq_flow_cell.path.parent
    )

    # GIVEN that the sample sheet is not in Housekeeper

    # WHEN testing if the sample sheet is in Housekeeper
    result: bool = demux_api.is_sample_sheet_in_housekeeper(flow_cell_id=tmp_bcl2fastq_flow_cell.id)

    # THEN the sample sheet should be in Housekeeper
    assert not result


def test_create_demultiplexing_output_dir_for_bcl2fastq(
    tmp_bcl2fastq_flow_cell: FlowCellDirectoryData,
    tmp_path: Path,
    demultiplexing_api: DemultiplexingAPI,
):
    """Test that the correct demultiplexing output directory is created."""
    # GIVEN a Bcl2Fastq FlowCellDirectoryData object

    # GIVEN that the demultiplexing output directory does not exist
    demultiplexing_api.demultiplexed_runs_dir = tmp_path
    output_directory: Path = demultiplexing_api.flow_cell_out_dir_path(tmp_bcl2fastq_flow_cell)
    unaligned_dir: Path = demultiplexing_api.get_flow_cell_unaligned_dir(tmp_bcl2fastq_flow_cell)

    assert not output_directory.exists()
    assert not unaligned_dir.exists()

    # WHEN creating the demultiplexing output directory for a bcl2fastq flow cell
    demultiplexing_api.create_demultiplexing_output_dir(tmp_bcl2fastq_flow_cell)

    # THEN the demultiplexing output directory should exist
    assert output_directory.exists()
    assert unaligned_dir.exists()


def test_create_demultiplexing_output_dir_for_bcl_convert(
    tmp_bcl_convert_flow_cell: FlowCellDirectoryData,
    tmp_path: Path,
    demultiplexing_api: DemultiplexingAPI,
):
    """Test that the correct demultiplexing output directory is created."""
    # GIVEN BCL Convert FlowCellDirectoryData object

    # GIVEN that the demultiplexing output directory does not exist
    demultiplexing_api.demultiplexed_runs_dir = tmp_path
    output_directory: Path = demultiplexing_api.flow_cell_out_dir_path(tmp_bcl_convert_flow_cell)
    unaligned_directory: Path = demultiplexing_api.get_flow_cell_unaligned_dir(
        tmp_bcl_convert_flow_cell
    )
    assert not output_directory.exists()
    assert not unaligned_directory.exists()

    # WHEN creating the demultiplexing output directory for a BCL Convert flow cell
    demultiplexing_api.create_demultiplexing_output_dir(tmp_bcl_convert_flow_cell)

    # THEN the demultiplexing output directory should exist
    assert output_directory.exists()
    assert not unaligned_directory.exists()


def test_is_demultiplexing_possible_true(
    demultiplexing_api: DemultiplexingAPI,
    tmp_bcl_convert_flow_cell: FlowCellDirectoryData,
):
    """Test demultiplexing pre-check when all criteria are fulfilled."""
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=tmp_bcl_convert_flow_cell.path,
        flow_cell_name=tmp_bcl_convert_flow_cell.id,
        hk_api=demultiplexing_api.hk_api,
    )
    # GIVEN a flow cell with no missing file

    # WHEN checking if demultiplexing is possible
    result: bool = demultiplexing_api.is_demultiplexing_possible(
        flow_cell=tmp_bcl_convert_flow_cell
    )
    # THEN the flow cell is ready for demultiplexing
    assert result is True


@pytest.mark.parametrize("missing_file", ["RTAComplete.txt", "CopyComplete.txt", "SampleSheet.csv"])
def test_is_demultiplexing_possible_missing_files(
    demultiplexing_api: DemultiplexingAPI,
    missing_file: str,
    tmp_bcl_convert_flow_cell: FlowCellDirectoryData,
):
    """Test demultiplexing pre-check when files are missing in flow cell directory."""
    # GIVEN a flow cell with a sample sheet in Housekeeper
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=tmp_bcl_convert_flow_cell.path,
        flow_cell_name=tmp_bcl_convert_flow_cell.id,
        hk_api=demultiplexing_api.hk_api,
    )

    # GIVEN that all other demultiplexing criteria are fulfilled
    assert (
        demultiplexing_api.is_demultiplexing_possible(flow_cell=tmp_bcl_convert_flow_cell) is True
    )

    # GIVEN a flow cell with a missing file
    Path(tmp_bcl_convert_flow_cell.path, missing_file).unlink()

    # WHEN checking if demultiplexing is possible
    result: bool = demultiplexing_api.is_demultiplexing_possible(
        flow_cell=tmp_bcl_convert_flow_cell
    )
    # THEN the flow cell should not be deemed ready for demultiplexing
    assert result is False


def is_demultiplexing_possible_no_sample_sheet_in_hk(
    demultiplexing_api: DemultiplexingAPI,
    tmp_bcl_convert_flow_cell: FlowCellDirectoryData,
):
    """Test demultiplexing pre-check when no sample sheet exists in Housekeeper."""
    # GIVEN a flow cell with no sample sheet in Housekeeper
    assert (
        demultiplexing_api.is_sample_sheet_in_housekeeper(flow_cell_id=tmp_bcl_convert_flow_cell.id)
        is False
    )

    # WHEN checking if demultiplexing is possible
    result: bool = demultiplexing_api.is_demultiplexing_possible(
        flow_cell=tmp_bcl_convert_flow_cell
    )
    # THEN the flow cell should not be deemed ready for demultiplexing
    assert result is False


def test_is_demultiplexing_possible_already_started(
    demultiplexing_api: DemultiplexingAPI,
    tmp_bcl_convert_flow_cell: FlowCellDirectoryData,
):
    """Test demultiplexing pre-check demultiplexing has already started."""
    # GIVEN a flow cell with a sample sheet in Housekeeper
    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=tmp_bcl_convert_flow_cell.path,
        flow_cell_name=tmp_bcl_convert_flow_cell.id,
        hk_api=demultiplexing_api.hk_api,
    )
    # GIVEN that all other demultiplexing criteria are fulfilled
    assert (
        demultiplexing_api.is_demultiplexing_possible(flow_cell=tmp_bcl_convert_flow_cell) is True
    )

    # GIVEN a flow cell where demultiplexing has already started
    Path(tmp_bcl_convert_flow_cell.path, DemultiplexingDirsAndFiles.DEMUX_STARTED).touch()

    # WHEN checking if demultiplexing is possible
    result: bool = demultiplexing_api.is_demultiplexing_possible(
        flow_cell=tmp_bcl_convert_flow_cell
    )
    # THEN the flow cell should not be deemed ready for demultiplexing
    assert result is False


def test_remove_demultiplexing_output_directory(
    demultiplexing_api: DemultiplexingAPI,
    tmp_path: Path,
    novaseq_6000_post_1_5_kits_flow_cell: FlowCellDirectoryData,
):
    """Test that the demultiplexing output directory is removed."""
    # GIVEN a flow cell with a demultiplexing output directory
    demultiplexing_api.demultiplexed_runs_dir = tmp_path
    demultiplexing_api.create_demultiplexing_output_dir(novaseq_6000_post_1_5_kits_flow_cell)
    assert demultiplexing_api.flow_cell_out_dir_path(novaseq_6000_post_1_5_kits_flow_cell).exists()

    # WHEN removing the demultiplexing output directory
    demultiplexing_api.remove_demultiplexing_output_directory(
        flow_cell=novaseq_6000_post_1_5_kits_flow_cell
    )

    assert not demultiplexing_api.flow_cell_out_dir_path(
        novaseq_6000_post_1_5_kits_flow_cell
    ).exists()



"""Tests for the prepare_fastq_api"""

from unittest import mock

import pytest

from cg.apps.crunchy import CrunchyAPI, crunchy
from cg.meta.compress import files
from cg.meta.compress.compress import CompressAPI
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI
from cg.models import CompressionData
from cg.store.models import Case
from cg.store.store import Store


def test_is_spring_decompression_needed_when_true(
    populated_compress_spring_api: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
    sample_id: str,
):
    """Test when spring decompression is needed."""

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_spring_api
    )
    # GIVEN a store with a case that has linked samples
    case_obj: Case = analysis_store_single_case.get_case_by_internal_id(internal_id=case_id)
    assert case_obj
    # GIVEN that the case has linked samples
    link_objects = [link_obj for link_obj in case_obj.links]
    assert link_objects
    # GIVEN a that there exists a version with only spring in housekeeper
    version_object = populated_compress_spring_api.hk_api.get_latest_bundle_version(
        bundle_name=sample_id
    )
    for file in version_object.files:
        assert file.path.endswith(".spring")

    # WHEN checking if spring decompression is needed
    res = prepare_fastq_api.is_spring_decompression_needed(case_id)

    # THEN assert that spring decompression is needed since there are no fastq files
    assert res is True


def test_is_spring_decompression_needed_when_false(
    populated_compress_api_fastq_spring: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
):
    """Test when spring decompression is not needed"""

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_api_fastq_spring
    )
    # GIVEN a store with a case that has linked samples
    case_obj: Case = analysis_store_single_case.get_case_by_internal_id(internal_id=case_id)
    assert case_obj
    # GIVEN that the case has linked samples
    link_objects = [link_obj for link_obj in case_obj.links]
    assert link_objects

    # WHEN checking if spring decompression is needed
    res = prepare_fastq_api.is_spring_decompression_needed(case_id)

    # THEN assert that spring decompression is not needed since there are fastq files
    assert res is False


def test_at_least_one_sample_be_decompressed(
    populated_compress_spring_api: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
    mocker,
):
    # GIVEN spring decompression is possible
    mocker.patch.object(CrunchyAPI, "is_spring_decompression_possible")
    CrunchyAPI.is_spring_decompression_possible.return_value = True

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_spring_api
    )

    # WHEN checking if spring decompression is possible
    res = prepare_fastq_api.can_at_least_one_sample_be_decompressed(case_id=case_id)

    # THEN assert that spring decompression is possible
    assert res is True


def test_no_samples_can_be_decompressed(
    populated_compress_spring_api: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
    mocker,
):
    # GIVEN spring decompression is not possible
    mocker.patch.object(CrunchyAPI, "is_spring_decompression_possible")
    CrunchyAPI.is_spring_decompression_possible.return_value = False

    # GIVEN a populated prepare_fastq_api
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_spring_api
    )

    # WHEN checking if spring decompression is possible
    res = prepare_fastq_api.can_at_least_one_sample_be_decompressed(case_id=case_id)

    # THEN assert that spring decompression is not possible
    assert res is False


@pytest.mark.parametrize("file_exists_and_is_accessible", [False, True])
def test_fastq_should_be_added_to_housekeeper(
    populated_compress_spring_api: CompressAPI,
    analysis_store_single_case: Store,
    case_id: str,
    sample_id: str,
    file_exists_and_is_accessible: bool,
):
    """Test when FASTQ needs to be added to Housekeeper."""

    # GIVEN a populated prepare_fastq_api and a case that has linked samples
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store_single_case, compress_api=populated_compress_spring_api
    )

    # GIVEN that there exists nr_of_files_before files in Housekeeper before adding any
    version_object = populated_compress_spring_api.hk_api.get_latest_bundle_version(
        bundle_name=sample_id
    )
    nr_of_files_before: int = len(list(version_object.files))

    # GIVEN that the decompressed files exist
    with mock.patch.object(
        CompressionData, "file_exists_and_is_accessible", return_value=file_exists_and_is_accessible
    ), mock.patch.object(crunchy.files, "get_crunchy_metadata", returnvalue=[]):
        # WHEN adding decompressed fastq files
        prepare_fastq_api.add_decompressed_fastq_files_to_housekeeper(case_id)

    # THEN assert that fastq files were added
    version_object = populated_compress_spring_api.hk_api.get_latest_bundle_version(
        bundle_name=sample_id
    )

    if file_exists_and_is_accessible:
        assert len(list(version_object.files)) > nr_of_files_before
    else:
        assert len(list(version_object.files)) == nr_of_files_before


def test_add_decompressed_fastqs_loops_through_samples(
    case_id: str, analysis_store: Store, populated_compress_spring_api: CompressAPI
):
    """Tests that given a case id, add_decompressed_fastq_files_to_housekeeper loops through all related samples."""

    # GIVEN a case with three samples and a PrepareFastqAPI
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store, compress_api=populated_compress_spring_api
    )
    with mock.patch.object(
        PrepareFastqAPI, "add_decompressed_sample", return_value=True
    ) as request_counter:
        # WHEN adding decompressed fastq files to Housekeeper
        prepare_fastq_api.add_decompressed_fastq_files_to_housekeeper(case_id)

    case = analysis_store.get_case_by_internal_id(case_id)

    # THEN each sample should have had their fastq files added to Housekeeper.
    assert request_counter.call_count == len(case.samples)


def test_add_decompressed_sample_loops_through_spring(
    case_id: str,
    analysis_store: Store,
    populated_compress_spring_api: CompressAPI,
    spring_file,
):
    """Tests that given a sample id, all fastq files are added to Housekeeper
    in add_decompressed_sample_to_housekeeper."""

    # GIVEN a case and one of its samples, which in turn has a spring file in Housekeeper but no fastq files
    prepare_fastq_api = PrepareFastqAPI(
        store=analysis_store, compress_api=populated_compress_spring_api
    )
    case = analysis_store.get_case_by_internal_id(case_id)
    sample = case.samples[0]

    with mock.patch.object(
        files,
        "get_hk_files_dict",
        return_value={},
    ), mock.patch.object(
        files,
        "get_spring_paths",
        return_value=[CompressionData(spring_file.with_suffix(""))],
    ), mock.patch.object(
        CompressAPI, "add_decompressed_fastq", return_value=True
    ) as request_submitter:
        # WHEN adding decompressed fastq files to Housekeeper
        prepare_fastq_api.add_decompressed_sample(case=case, sample=sample)

    # THEN the related fastq files should have been added to Housekeeper
    assert request_submitter.call_count == 1



from sqlalchemy.orm import Query

from cg.store.filters.status_bed_filters import (
    get_bed_by_entry_id,
    get_bed_by_name,
    get_not_archived_beds,
    order_beds_by_name,
)
from cg.store.models import Bed
from cg.store.store import Store


def test_get_bed_by_entry_id(base_store: Store, entry_id: int = 1):
    """Test return panel bed by entry id."""
    # GIVEN a store containing bed

    # WHEN retrieving a bed by id
    bed: Bed = get_bed_by_entry_id(
        beds=base_store._get_query(table=Bed), bed_entry_id=entry_id
    ).first()

    # THEN panel bed should be returned
    assert bed

    # THEN the filename should match the original
    assert bed.id == entry_id


def test_get_bed_by_entry_id_no_id(base_store: Store):
    """Test return panel bed by entry id when invalid id."""
    # GIVEN a store containing bed

    # WHEN retrieving a bed by an invalid id
    bed: Bed = get_bed_by_entry_id(beds=base_store._get_query(table=Bed), bed_entry_id=999).first()

    # THEN bed should not be returned
    assert not bed


def test_get_bed_by_name(base_store: Store, bed_name: str):
    """Test return bed by name."""
    # GIVEN a store containing bed

    # WHEN retrieving a bed by name
    bed: Bed = get_bed_by_name(beds=base_store._get_query(table=Bed), bed_name=bed_name).first()

    # THEN panel bed should be returned
    assert bed

    # THEN the name should match the original
    assert bed.name == bed_name


def test_get_bed_by_name_no_name(base_store: Store):
    """Test return bed by name when no valid name is provided."""
    # GIVEN a store containing bed

    # WHEN retrieving a bed by name
    bed: Bed = get_bed_by_name(beds=base_store._get_query(table=Bed), bed_name="not_a_name").first()

    # THEN panel bed should not be returned
    assert not bed


def test_get_not_archived_beds(base_store: Store):
    """Test return bed not archived."""
    # GIVEN a store containing bed

    # WHEN retrieving a beds not archived
    beds: Query = get_not_archived_beds(beds=base_store._get_query(table=Bed))

    # ASSERT that the beds is a query
    assert isinstance(beds, Query)

    # THEN beds should be returned
    assert beds

    # THEN the bed attribute is_archived should be False
    assert beds[0].is_archived is False


def test_order_beds_by_name(base_store: Store, bed_name: str):
    """Test return bed ordered by name."""
    # GIVEN a store containing bed

    # WHEN retrieving a beds
    beds: Query = order_beds_by_name(beds=base_store._get_query(table=Bed))

    # ASSERT that the beds is a query
    assert isinstance(beds, Query)

    # THEN a bed should be returned
    assert beds

    # THEN the bed attribute name should match the original
    assert beds[0].name == bed_name



import os
import shutil
from collections import namedtuple
from datetime import datetime
from pathlib import Path

import pytest

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.demultiplexing import BclConverter, DemultiplexingDirsAndFiles
from cg.meta.demultiplex.demux_post_processing import DemuxPostProcessingAPI
from cg.meta.demultiplex.housekeeper_storage_functions import (
    add_sample_sheet_path_to_housekeeper,
)
from cg.models.cg_config import CGConfig
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.store.models import Case, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers

FlowCellInfo = namedtuple("FlowCellInfo", "directory name sample_internal_ids")


@pytest.fixture(name="tmp_demulitplexing_dir")
def tmp_illumina_novaseq_demulitplexing_dir(
    illumina_demultiplexed_runs_directory: Path, bcl2fastq_flow_cell_full_name: str
) -> Path:
    """Return a tmp directory in demultiplexed-runs."""
    tmp_demulitplexing_dir: Path = Path(
        illumina_demultiplexed_runs_directory, bcl2fastq_flow_cell_full_name
    )
    tmp_demulitplexing_dir.mkdir(exist_ok=True, parents=True)
    return tmp_demulitplexing_dir


@pytest.fixture(name="tmp_fastq_paths")
def temp_fastq_paths(tmp_demulitplexing_dir: Path) -> list[Path]:
    """Return a list of temporary dummy fastq paths."""
    fastqs = [
        Path(tmp_demulitplexing_dir, "fastq_1.fastq.gz"),
        Path(tmp_demulitplexing_dir, "fastq_2.fastq.gz"),
    ]
    for fastq in fastqs:
        with fastq.open("w+") as fh:
            fh.write("content")
    return fastqs


@pytest.fixture
def demultiplex_fastq_file_path() -> Path:
    """Return a path to non-existent a fastq file."""
    return Path("path/to/sample_internal_id_S1_L001_R1_001.fastq.gz")


@pytest.fixture(name="tmp_sample_sheet_path")
def tmp_samplesheet_path(tmp_demulitplexing_dir: Path) -> Path:
    """Return SampleSheet in temporary demuliplexing folder."""
    tmp_sample_sheet_path = Path(tmp_demulitplexing_dir, "SampleSheet.csv")
    with tmp_sample_sheet_path.open("w+") as fh:
        fh.write("content")
    return tmp_sample_sheet_path


@pytest.fixture(name="tmp_flow_cell_run_base_path")
def tmp_flow_cell_run_base_path(project_dir: Path, bcl2fastq_flow_cell_full_name: str) -> Path:
    """Flow cell run directory in temporary folder."""

    tmp_flow_cell_run_path: Path = Path(project_dir, "flow_cells")
    tmp_flow_cell_run_path.mkdir(exist_ok=True, parents=True)

    return tmp_flow_cell_run_path


@pytest.fixture(name="tmp_flow_cell_demux_base_path")
def tmp_flow_cell_demux_base_path(project_dir: Path, bcl2fastq_flow_cell_full_name: str) -> Path:
    """Flow cell demux directory in temporary folder."""

    tmp_flow_cell_demux_path: Path = Path(project_dir, "demultiplexed-runs")
    tmp_flow_cell_demux_path.mkdir(exist_ok=True, parents=True)

    return tmp_flow_cell_demux_path


@pytest.fixture(name="flow_cell_project_id")
def flow_cell_project_id() -> int:
    """Return flow cell run project id."""
    return 174578


@pytest.fixture(name="populated_flow_cell_store")
def populated_flow_cell_store(
    family_name: str,
    bcl2fastq_flow_cell_id: str,
    sample_id: str,
    store: Store,
    helpers: StoreHelpers,
) -> Store:
    """Populate a store with a NovaSeq flow cell."""

    populated_flow_cell_store: Store = store
    sample: Sample = helpers.add_sample(store=populated_flow_cell_store, internal_id=sample_id)
    family: Case = helpers.add_case(store=populated_flow_cell_store, internal_id=family_name)
    helpers.add_relationship(
        store=populated_flow_cell_store,
        sample=sample,
        case=family,
    )
    helpers.add_flow_cell(
        store=populated_flow_cell_store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        sequencer_type="novaseq",
        samples=[sample],
    )
    return populated_flow_cell_store


@pytest.fixture(name="active_flow_cell_store")
def active_flow_cell_store(
    family_name: str,
    bcl2fastq_flow_cell_id: str,
    sample_id: str,
    base_store: Store,
    helpers: StoreHelpers,
) -> Store:
    """Populate a store with a Novaseq flow cell, with active samples on it."""
    active_flow_cell_store: Store = base_store
    sample: Sample = helpers.add_sample(store=active_flow_cell_store, internal_id=sample_id)
    family: Case = helpers.add_case(
        store=active_flow_cell_store, internal_id=family_name, action="running"
    )
    helpers.add_relationship(
        store=active_flow_cell_store,
        sample=sample,
        case=family,
    )
    helpers.add_flow_cell(
        store=active_flow_cell_store,
        flow_cell_name=bcl2fastq_flow_cell_id,
        sequencer_type="novaseq",
        samples=[sample],
    )
    return active_flow_cell_store


@pytest.fixture(name="sample_level_housekeeper_api")
def sample_level_housekeeper_api(
    bcl2fastq_flow_cell_id: str,
    real_housekeeper_api: HousekeeperAPI,
    sample_id: str,
    tmp_fastq_paths: list[Path],
    helpers,
) -> HousekeeperAPI:
    """Return a mocked Housekeeper API, containing a sample bundle with related FASTQ files."""
    sample_level_housekeeper_api = real_housekeeper_api
    bundle_data = {
        "name": sample_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {"path": path.as_posix(), "tags": ["fastq", bcl2fastq_flow_cell_id], "archive": False}
            for path in tmp_fastq_paths
        ],
    }
    helpers.ensure_hk_bundle(store=sample_level_housekeeper_api, bundle_data=bundle_data)
    return sample_level_housekeeper_api


@pytest.fixture(name="flow_cell_name_housekeeper_api")
def flow_cell_name_housekeeper_api(
    bcl2fastq_flow_cell_id: str,
    real_housekeeper_api: HousekeeperAPI,
    sample_id: str,
    tmp_fastq_paths: list[Path],
    tmp_sample_sheet_path: Path,
    helpers,
) -> HousekeeperAPI:
    """Return a mocked Housekeeper API, containing a sample bundle with related FASTQ files."""
    flow_cell_housekeeper_api = real_housekeeper_api
    bundle_data = {
        "name": sample_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {"path": path.as_posix(), "tags": ["fastq", bcl2fastq_flow_cell_id], "archive": False}
            for path in tmp_fastq_paths
        ],
    }
    flow_cell_bundle_data = {
        "name": bcl2fastq_flow_cell_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {
                "path": tmp_sample_sheet_path.as_posix(),
                "tags": ["samplesheet", bcl2fastq_flow_cell_id],
                "archive": False,
            }
        ],
    }

    helpers.ensure_hk_bundle(store=flow_cell_housekeeper_api, bundle_data=bundle_data)
    helpers.ensure_hk_bundle(store=flow_cell_housekeeper_api, bundle_data=flow_cell_bundle_data)
    return flow_cell_housekeeper_api


@pytest.fixture(scope="session")
def flow_cell_name_demultiplexed_with_bcl_convert() -> str:
    return "HY7FFDRX2"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl_convert(
    flow_cell_name_demultiplexed_with_bcl_convert: str,
):
    return f"230504_A00689_0804_B{flow_cell_name_demultiplexed_with_bcl_convert}"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl_convert_flat(
    flow_cell_name_demultiplexed_with_bcl_convert: str,
):
    """Return the name of a flow cell directory that has been demultiplexed with Bcl Convert using a flat output directory structure."""
    return f"230505_A00689_0804_B{flow_cell_name_demultiplexed_with_bcl_convert}"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl_convert_on_sequencer(
    flow_cell_name_demultiplexed_with_bcl_convert_on_sequencer: str,
):
    """Return the name of a flow cell directory that has been demultiplexed with Bcl Convert on the NovaseqX sequencer."""
    return f"20230508_LH00188_0003_A{flow_cell_name_demultiplexed_with_bcl_convert_on_sequencer}"


@pytest.fixture(scope="session")
def flow_cell_name_demultiplexed_with_bcl_convert_on_sequencer() -> str:
    """Return the name of a flow cell directory that has been demultiplexed with Bcl Convert on the NovaseqX sequencer."""
    return "22522YLT3"


@pytest.fixture(scope="session")
def bcl2fastq_folder_structure(tmp_path_factory, cg_dir: Path) -> Path:
    """Return a folder structure that resembles a bcl2fastq run folder."""
    base_dir: Path = tmp_path_factory.mktemp("".join((str(cg_dir), BclConverter.BCL2FASTQ)))
    folders: list[str] = ["l1t21", "l1t11", "l2t11", "l2t21"]

    for folder in folders:
        new_dir: Path = Path(base_dir, folder)
        new_dir.mkdir()

    return base_dir


@pytest.fixture(scope="function")
def not_bcl2fastq_folder_structure(tmp_path_factory, cg_dir: Path) -> Path:
    """Return a folder structure that does not resemble a bcl2fastq run folder."""
    base_dir: Path = tmp_path_factory.mktemp("".join((str(cg_dir), "not_bcl2fastq")))
    folders: list[str] = ["just", "some", "folders"]

    for folder in folders:
        new_dir: Path = Path(base_dir, folder)
        new_dir.mkdir()

    return base_dir


@pytest.fixture(scope="session")
def base_call_file() -> Path:
    return Path("Data", "Intensities", "BaseCalls", "L001", "C1.1", "L001_1.cbcl")


@pytest.fixture(scope="session")
def inter_op_file() -> Path:
    return Path(DemultiplexingDirsAndFiles.INTER_OP, "AlignmentMetricsOut.bin")


@pytest.fixture(scope="session")
def thumbnail_file() -> Path:
    return Path("Thumbnail_Images", "L001", "C1.1", "s_1_1105_green.png")


@pytest.fixture
def lsyncd_source_directory(
    tmp_path_factory,
    novaseq_x_manifest_file: Path,
    base_call_file: Path,
    inter_op_file: Path,
    thumbnail_file: Path,
) -> Path:
    """Return a temporary directory with a manifest file and three dummy files."""
    source_directory = Path(tmp_path_factory.mktemp("source"))
    shutil.copy(novaseq_x_manifest_file, source_directory)
    for file in [base_call_file, inter_op_file, thumbnail_file]:
        full_path = Path(source_directory, file)
        full_path.parent.mkdir(parents=True)
        full_path.touch()
    return source_directory


@pytest.fixture
def lsyncd_target_directory(lsyncd_source_directory: Path, tmp_path_factory) -> Path:
    """Return a copy of the temporary source directory."""
    temp_target_directory = Path(tmp_path_factory.mktemp("tmp_target"))
    target_directory = Path(lsyncd_source_directory.parent, Path(temp_target_directory, "target"))
    shutil.copytree(lsyncd_source_directory, target_directory)
    return target_directory


@pytest.fixture
def demux_post_processing_api(
    demultiplex_context: CGConfig, tmp_illumina_demultiplexed_flow_cells_directory
) -> DemuxPostProcessingAPI:
    api = DemuxPostProcessingAPI(demultiplex_context)
    api.demultiplexed_runs_dir = tmp_illumina_demultiplexed_flow_cells_directory
    return api


@pytest.fixture
def bcl2fastq_flow_cell_dir_name(demux_post_processing_api) -> str:
    """Return a flow cell name that has been demultiplexed with bcl2fastq."""
    flow_cell_dir_name = "170407_ST-E00198_0209_BHHKVCALXX"
    flow_cell_path = Path(demux_post_processing_api.demultiplexed_runs_dir, flow_cell_dir_name)

    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=flow_cell_path,
        flow_cell_name="HHKVCALXX",
        hk_api=demux_post_processing_api.hk_api,
    )
    return flow_cell_dir_name


@pytest.fixture
def bcl2fastq_sample_id_with_non_pooled_undetermined_reads() -> str:
    return "SVE2528A1"


@pytest.fixture
def bcl2fastq_non_pooled_sample_read_count() -> int:
    """Based on the data in 170407_ST-E00198_0209_BHHKVCALXX, the sum of all reads - mapped and undetermined."""
    return 8000000


@pytest.fixture
def bclconvert_flow_cell_dir_name(demux_post_processing_api) -> str:
    """Return a flow cell name that has been demultiplexed with bclconvert."""
    flow_cell_dir_name = "230504_A00689_0804_BHY7FFDRX2"
    flow_cell_path = Path(demux_post_processing_api.demultiplexed_runs_dir, flow_cell_dir_name)

    add_sample_sheet_path_to_housekeeper(
        flow_cell_directory=flow_cell_path,
        flow_cell_name="HY7FFDRX2",
        hk_api=demux_post_processing_api.hk_api,
    )
    return flow_cell_dir_name


@pytest.fixture
def bcl_convert_sample_id_with_non_pooled_undetermined_reads() -> str:
    return "ACC11927A2"


@pytest.fixture
def bcl_convert_non_pooled_sample_read_count() -> int:
    """Based on the data in 230504_A00689_0804_BHY7FFDRX2, the sum of all reads - mapped and undetermined."""
    return 4000000


def get_all_files_in_directory_tree(directory: Path) -> list[Path]:
    """Get the relative paths of all files in a directory and its subdirectories."""
    files_in_directory: list[Path] = []
    for subdir, _, files in os.walk(directory):
        subdir = Path(subdir).relative_to(directory)
        files_in_directory.extend([Path(subdir, file) for file in files])
    return files_in_directory



from pydantic.v1 import BaseModel, constr, validator

from cg.constants import DataDelivery
from cg.constants.constants import GenomeVersion, Workflow
from cg.models.orders.order import OrderType
from cg.models.orders.sample_base import (
    NAME_PATTERN,
    ContainerEnum,
    ControlEnum,
    PriorityEnum,
    SexEnum,
    StatusEnum,
)
from cg.store.models import Application, Case, Organism, Panel, Pool, Sample


class OptionalIntValidator:
    @classmethod
    def str_to_int(cls, v: str) -> int | None:
        return int(v) if v else None


class OptionalFloatValidator:
    @classmethod
    def str_to_float(cls, v: str) -> float | None:
        return float(v) if v else None


class OrderInSample(BaseModel):
    # Order portal specific
    internal_id: constr(max_length=Sample.internal_id.property.columns[0].type.length) | None
    _suitable_project: OrderType = None
    application: constr(max_length=Application.tag.property.columns[0].type.length)
    comment: constr(max_length=Sample.comment.property.columns[0].type.length) | None
    skip_reception_control: bool | None = None
    data_analysis: Workflow
    data_delivery: DataDelivery
    name: constr(
        regex=NAME_PATTERN,
        min_length=2,
        max_length=Sample.name.property.columns[0].type.length,
    )
    priority: PriorityEnum = PriorityEnum.standard
    require_qc_ok: bool = False
    volume: str
    concentration_ng_ul: str | None

    @classmethod
    def is_sample_for(cls, project: OrderType):
        return project == cls._suitable_project


class Of1508Sample(OrderInSample):
    # Orderform 1508
    # Order portal specific
    internal_id: constr(max_length=Sample.internal_id.property.columns[0].type.length) | None
    # "required for new samples"
    name: (
        constr(
            regex=NAME_PATTERN,
            min_length=2,
            max_length=Sample.name.property.columns[0].type.length,
        )
        | None
    )

    # customer
    age_at_sampling: float | None
    family_name: constr(
        regex=NAME_PATTERN,
        min_length=2,
        max_length=Case.name.property.columns[0].type.length,
    )
    case_internal_id: constr(max_length=Sample.internal_id.property.columns[0].type.length) | None
    sex: SexEnum = SexEnum.unknown
    tumour: bool = False
    source: str | None
    control: ControlEnum | None
    volume: str | None
    container: ContainerEnum | None
    # "required if plate for new samples"
    container_name: str | None
    well_position: str | None
    # "Required if samples are part of trio/family"
    mother: (
        constr(regex=NAME_PATTERN, max_length=Sample.name.property.columns[0].type.length) | None
    )
    father: (
        constr(regex=NAME_PATTERN, max_length=Sample.name.property.columns[0].type.length) | None
    )
    # This information is required for panel analysis
    capture_kit: str | None
    # This information is required for panel- or exome analysis
    elution_buffer: str | None
    tumour_purity: int | None
    # "This information is optional for FFPE-samples for new samples"
    formalin_fixation_time: int | None
    post_formalin_fixation_time: int | None
    tissue_block_size: str | None
    # "Not Required"
    cohorts: list[str] | None
    phenotype_groups: list[str] | None
    phenotype_terms: list[str] | None
    require_qc_ok: bool = False
    quantity: int | None
    subject_id: (
        constr(regex=NAME_PATTERN, max_length=Sample.subject_id.property.columns[0].type.length)
        | None
    )
    synopsis: str | None

    @validator("container", "container_name", "name", "source", "subject_id", "volume")
    def required_for_new_samples(cls, value, values, **kwargs):
        if not value and not values.get("internal_id"):
            raise ValueError(f"required for new sample {values.get('name')}")
        return value

    @validator(
        "tumour_purity",
        "formalin_fixation_time",
        "post_formalin_fixation_time",
        "quantity",
        pre=True,
    )
    def str_to_int(cls, v: str) -> int | None:
        return OptionalIntValidator.str_to_int(v=v)

    @validator(
        "age_at_sampling",
        "volume",
        pre=True,
    )
    def str_to_float(cls, v: str) -> float | None:
        return OptionalFloatValidator.str_to_float(v=v)


class MipDnaSample(Of1508Sample):
    _suitable_project = OrderType.MIP_DNA
    # "Required if data analysis in Scout or vcf delivery"
    panels: list[constr(min_length=1, max_length=Panel.abbrev.property.columns[0].type.length)]
    status: StatusEnum


class BalsamicSample(Of1508Sample):
    _suitable_project = OrderType.BALSAMIC


class BalsamicQCSample(Of1508Sample):
    _suitable_project = OrderType.BALSAMIC_QC
    reference_genome: GenomeVersion | None


class BalsamicUmiSample(Of1508Sample):
    _suitable_project = OrderType.BALSAMIC_UMI


class MipRnaSample(Of1508Sample):
    _suitable_project = OrderType.MIP_RNA


class RnafusionSample(Of1508Sample):
    _suitable_project = OrderType.RNAFUSION


class FastqSample(OrderInSample):
    _suitable_project = OrderType.FASTQ

    # Orderform 1508
    # "required"
    container: ContainerEnum | None
    sex: SexEnum = SexEnum.unknown
    source: str
    tumour: bool
    # "required if plate"
    container_name: str | None
    well_position: str | None
    elution_buffer: str
    # This information is required for panel analysis
    capture_kit: str | None
    # "Not Required"
    quantity: int | None
    subject_id: str | None

    @validator("quantity", pre=True)
    def str_to_int(cls, v: str) -> int | None:
        return OptionalIntValidator.str_to_int(v=v)


class RmlSample(OrderInSample):
    _suitable_project = OrderType.RML

    # 1604 Orderform Ready made libraries (RML)
    # Order portal specific
    # "This information is required"
    pool: constr(max_length=Pool.name.property.columns[0].type.length)
    concentration: float
    concentration_sample: float | None
    index: str
    index_number: str | None
    # "Required if Plate"
    rml_plate_name: str | None
    well_position_rml: str | None
    # "Automatically generated (if not custom) or custom"
    index_sequence: str | None
    # "Not required"
    control: str | None

    @validator("concentration_sample", pre=True)
    def str_to_float(cls, v: str) -> float | None:
        return OptionalFloatValidator.str_to_float(v=v)


class FluffySample(RmlSample):
    _suitable_project = OrderType.FLUFFY
    # 1604 Orderform Ready made libraries (RML)


class MetagenomeSample(Of1508Sample):
    _suitable_project = OrderType.METAGENOME
    # "This information is required"
    source: str
    # "This information is not required"
    concentration_sample: float | None
    family_name: None = None
    subject_id: None = None

    @validator("concentration_sample", pre=True)
    def str_to_float(cls, v: str) -> float | None:
        return OptionalFloatValidator.str_to_float(v=v)

    @validator("subject_id", pre=True)
    def required_for_new_samples(cls, v: str) -> None:
        """Overrides the parent validator since subject_id is optional for these samples."""
        return None


class MicrobialSample(OrderInSample):
    # 1603 Orderform Microbial WGS
    # "These fields are required"
    organism: constr(max_length=Organism.internal_id.property.columns[0].type.length)
    reference_genome: constr(max_length=Sample.reference_genome.property.columns[0].type.length)
    elution_buffer: str
    extraction_method: str | None
    container: ContainerEnum
    # "Required if Plate"
    container_name: str | None
    well_position: str | None
    # "Required if "Other" is chosen in column "Species""
    organism_other: constr(max_length=Organism.internal_id.property.columns[0].type.length) | None
    # "These fields are not required"
    concentration_sample: float | None
    quantity: int | None
    verified_organism: bool | None  # sent to LIMS
    control: str | None

    @validator("quantity", pre=True)
    def str_to_int(cls, v: str) -> int | None:
        return OptionalIntValidator.str_to_int(v=v)

    @validator("concentration_sample", pre=True)
    def str_to_float(cls, v: str) -> float | None:
        return OptionalFloatValidator.str_to_float(v=v)


class MicrosaltSample(MicrobialSample):
    _suitable_project = OrderType.MICROSALT
    # 1603 Orderform Microbial WGS


class SarsCov2Sample(MicrobialSample):
    _suitable_project = OrderType.SARS_COV_2

    # 2184 Orderform SARS-COV-2
    # "These fields are required"
    collection_date: str
    lab_code: str
    primer: str
    original_lab: str
    original_lab_address: str
    pre_processing_method: str
    region: str
    region_code: str
    selection_criteria: str
    volume: str | None


def sample_class_for(project: OrderType):
    """Get the sample class for the specified project

    Args:
        project     (OrderType):    Project to get sample subclass for
    Returns:
        Subclass of OrderInSample
    """

    def all_subclasses(cls):
        """Get all subclasses recursively for a class

        Args:
            cls     (Class):    Class to get all subclasses for
        Returns:
            Set of Subclasses of cls
        """
        if cls.__subclasses__():
            return set(cls.__subclasses__()).union(
                [s for c in cls.__subclasses__() for s in all_subclasses(c)]
            )

        return []

    for sub_cls in all_subclasses(OrderInSample):
        if sub_cls.is_sample_for(project):
            return sub_cls

    raise ValueError



import pytest

from cg.apps.orderform.orderform_parser import OrderformParser
from cg.exc import OrderFormError


class Shape:
    def __init__(self, name, sides):
        self.name = name
        self.sides = sides


def test__get_single_value_when_same():
    """Test method to get a single value"""

    # GIVEN a bunch of items with an attribute where all items have the same value
    expected_sides: int = 4
    items: [{}] = [
        Shape(name="rhombus", sides=expected_sides),
        Shape(name="square", sides=expected_sides),
    ]

    # WHEN getting a single value
    single_value: int = OrderformParser._get_single_value(
        items_id="items", items=items, attr="sides"
    )

    # THEN assert that the returned value was the expected one
    assert single_value == expected_sides


def test__get_single_value_when_different():
    """Test method to get a single value"""

    # GIVEN a bunch of items with an attribute where all items have different values
    items: [{}] = [Shape(name="ball", sides=1), Shape(name="square", sides=4)]

    # WHEN getting a single value but items contains different values
    # THEN assert that the expected exception is raised
    with pytest.raises(OrderFormError):
        OrderformParser._get_single_value(items_id="items", items=items, attr="sides")


def test__get_single_value_when_same():
    """Test method to get a single value"""

    # GIVEN a bunch of items with an attribute where all items have the same value
    expected_sides: [] = ["A", "B", "C", "D"]
    items: [{}] = [
        Shape(name="rhombus", sides=expected_sides),
        Shape(name="square", sides=expected_sides),
    ]

    # WHEN getting a single value
    single_value: [] = OrderformParser._get_single_set(items_id="items", items=items, attr="sides")

    # THEN assert that the returned value was the expected one
    assert single_value == set(expected_sides)


def test__get_single_set_when_different():
    """Test method to get a single value"""

    # GIVEN a bunch of items with an attribute where all items have different values
    items: [{}] = [
        Shape(name="ball", sides=["outside", "inside"]),
        Shape(name="square", sides=["r", "l", "u", "d"]),
    ]

    # WHEN getting a single value but items contains different values
    # THEN assert that the expected exception is raised
    with pytest.raises(OrderFormError):
        OrderformParser._get_single_set(items_id="items", items=items, attr="sides")



"""Tests for functions related to indexes."""

import pytest

from cg.apps.demultiplex.sample_sheet.index import (
    Index,
    get_hamming_distance_index_1,
    get_hamming_distance_index_2,
    get_reverse_complement_dna_seq,
    get_valid_indexes,
    is_padding_needed,
)


def test_get_valid_indexes():
    """Test that the function get_valid_indexes returns a list of Index objects."""
    # GIVEN a sample sheet API

    # WHEN fetching the indexes
    indexes: list[Index] = get_valid_indexes()

    # THEN assert that the indexes are correct
    assert indexes
    assert isinstance(indexes[0], Index)


@pytest.mark.parametrize(
    "index1_cycles, index2_cycles, sample_index_length, expected_is_padding_needed",
    [
        (10, 10, 8, True),
        (10, 10, 10, False),
        (17, 8, 17, False),
        (8, 8, 10, False),
        (8, 8, 8, False),
    ],
)
def test_is_padding_needed(
    index1_cycles: int,
    index2_cycles: int,
    sample_index_length: int,
    expected_is_padding_needed: bool,
):
    """Test that evaluating if a situation needs padding returns the expected value."""
    # GIVEN a sample index length and the number of index cycles reads stated in the run parameters

    # WHEN checking if padding is needed
    padding_needed: bool = is_padding_needed(
        index1_cycles=index1_cycles,
        index2_cycles=index2_cycles,
        sample_index_length=sample_index_length,
    )

    # THEN assert that the result is the expected
    assert padding_needed == expected_is_padding_needed


def test_get_reverse_complement():
    """Test that getting the reverse complement of a DNA strain returns the correct sequence."""
    # GIVEN a DNA strain and its reverse complement
    dna_strain: str = "ACCTCTGT"
    reversed_complement: str = "ACAGAGGT"

    # WHEN getting the reverse complement of the DNA strain
    returned_reverse_complement: str = get_reverse_complement_dna_seq(dna=dna_strain)

    # THEN the result is the expected
    assert returned_reverse_complement == reversed_complement


def test_get_reverse_complement_not_dna(caplog):
    """Test that using a string without 'A', 'C', 'G' or 'T' fails."""
    # GIVEN a non-DNA strain
    strain: str = "ACCUCTGU"

    # WHEN getting the reverse complement
    with pytest.raises(KeyError):
        # THEN the process fails due to not recognising the unknown nucleotide
        get_reverse_complement_dna_seq(dna=strain)


@pytest.mark.parametrize(
    "sequence_1, sequence_2, expected_distance",
    [
        ("GATTACA", "GATTACA", 0),
        ("GATTACA", "GATTACAXX", 0),
        ("XXXACA", "GATTACA", 6),
        ("XXXXXXX", "GATTACA", 7),
    ],
    ids=[
        "Identical sequences",
        "Same initial part, different lengths",
        "Same final part, different lengths",
        "Different sequences, same length",
    ],
)
def test_get_hamming_distance_index_1(sequence_1: str, sequence_2: str, expected_distance: int):
    """
    Test that Hamming distances are calculated correctly for different sets of index 1 sequences.
    This is, that the operation is commutative and aligns sequences from the left.
    """
    # GIVEN two index_1 sequences

    # WHEN getting the hamming distance between them in any order

    # THEN the distance is zero
    assert (
        get_hamming_distance_index_1(sequence_1=sequence_1, sequence_2=sequence_2)
        == expected_distance
    )
    assert (
        get_hamming_distance_index_1(sequence_1=sequence_2, sequence_2=sequence_1)
        == expected_distance
    )


@pytest.mark.parametrize(
    "sequence_1, sequence_2, expected_distance",
    [
        ("GATTACA", "GATTACA", 0),
        ("GATTACA", "XXGATTACA", 0),
        ("GATXX", "GATTACA", 5),
        ("XXXXXXX", "GATTACA", 7),
    ],
    ids=[
        "Identical sequences",
        "Same final part, different lengths",
        "Same initial part, different lengths",
        "Different sequences, same length",
    ],
)
def test_get_hamming_distance_index_2_reverse_complement(
    sequence_1: str, sequence_2: str, expected_distance: int
):
    """
    Test that Hamming distances are calculated correctly for different sets of index 2 sequences
    with reverse complement. This is, that the operation is commutative and aligns sequences from
    the right.
    """
    # GIVEN two index_2 sequences

    # WHEN getting the hamming distance between them in any order

    # THEN the distance is zero
    assert (
        get_hamming_distance_index_2(
            sequence_1=sequence_1, sequence_2=sequence_2, is_reverse_complement=True
        )
        == expected_distance
    )
    assert (
        get_hamming_distance_index_2(
            sequence_1=sequence_2, sequence_2=sequence_1, is_reverse_complement=True
        )
        == expected_distance
    )


@pytest.mark.parametrize(
    "sequence_1, sequence_2, expected_distance",
    [
        ("GATTACA", "GATTACA", 0),
        ("GATTACA", "GATTACAXX", 0),
        ("XXXACA", "GATTACA", 6),
        ("XXXXXXX", "GATTACA", 7),
    ],
    ids=[
        "Identical sequences",
        "Same initial part, different lengths",
        "Same final part, different lengths",
        "Different sequences, same length",
    ],
)
def test_get_hamming_distance_index_2_no_reverse_complement(
    sequence_1: str, sequence_2: str, expected_distance: int
):
    """
    Test that Hamming distances are calculated correctly for different sets of index 2 sequences
    without reverse complement. This is, that the operation is commutative and aligns sequences
    from the left.
    """
    # GIVEN two index_2 sequences

    # WHEN getting the hamming distance between them in any order with reverse complement

    # THEN the distance is zero
    assert (
        get_hamming_distance_index_2(
            sequence_1=sequence_1, sequence_2=sequence_2, is_reverse_complement=False
        )
        == expected_distance
    )
    assert (
        get_hamming_distance_index_2(
            sequence_1=sequence_2, sequence_2=sequence_1, is_reverse_complement=False
        )
        == expected_distance
    )



"""Tests for the compress data class"""

from datetime import datetime
from pathlib import Path

from cg.models import CompressionData


def test_get_run_name():
    """Test that the correct run name is returned"""
    # GIVEN a file path that ends with a run name
    file_path = Path("/path/to/dir")
    run_name = "a_run"
    # GIVEN a compression data object
    compression_obj = CompressionData(file_path / run_name)

    # WHEN fetching the run name
    # THEN assert the correct run name is returned
    assert compression_obj.run_name == run_name


def test_get_change_date(compression_object):
    """Test to get the date time for when a file was changed"""
    # GIVEN a existing file
    file_path = compression_object.spring_path
    file_path.touch()

    # WHEN fetching the date when file was created
    change_date = compression_object.get_change_date(file_path)

    # THEN check that it is the same date as today
    assert change_date.date() == datetime.today().date()



from cg.constants.constants import MicrosaltAppTags
from cg.meta.workflow.microsalt.metrics_parser.models import SampleMetrics
from cg.meta.workflow.microsalt.quality_controller.models import SampleQualityResult
from cg.meta.workflow.microsalt.quality_controller.utils import (
    get_non_urgent_results,
    get_urgent_results,
    has_valid_10x_coverage,
    has_valid_average_coverage,
    has_valid_duplication_rate,
    has_valid_mapping_rate,
    has_valid_median_insert_size,
    is_valid_10x_coverage,
    is_valid_average_coverage,
    is_valid_duplication_rate,
    is_valid_mapping_rate,
    is_valid_median_insert_size,
    is_valid_total_reads,
    is_valid_total_reads_for_negative_control,
    negative_control_pass_qc,
    non_urgent_samples_pass_qc,
    urgent_samples_pass_qc,
)
from tests.meta.workflow.microsalt.conftest import create_quality_result, create_sample_metrics


def test_sample_total_reads_passing():
    # GIVEN sufficient reads
    sample_reads = 100
    target_reads = 100

    # WHEN checking if the sample has sufficient reads
    passes_reads_threshold: bool = is_valid_total_reads(
        reads=sample_reads, target_reads=target_reads, threshold_percentage=90
    )

    # THEN it passes
    assert passes_reads_threshold


def test_sample_total_reads_failing():
    # GIVEN insufficient reads
    sample_reads = 50
    target_reads = 100

    # WHEN checking if the sample has sufficient reads
    passes_reads_threshold: bool = is_valid_total_reads(
        reads=sample_reads, target_reads=target_reads, threshold_percentage=90
    )

    # THEN it fails
    assert not passes_reads_threshold


def test_sample_total_reads_failing_without_reads():
    # GIVENout reads
    sample_reads = 0
    target_reads = 100

    # WHEN checking if the sample has sufficient reads
    passes_reads_threshold: bool = is_valid_total_reads(
        reads=sample_reads, target_reads=target_reads, threshold_percentage=90
    )

    # THEN it fails
    assert not passes_reads_threshold


def test_control_total_reads_passing():
    # GIVEN a negative control sample with few reads
    sample_reads = 1
    target_reads = 100

    # WHEN checking if the control read count is valid
    passes_reads_threshold: bool = is_valid_total_reads_for_negative_control(
        reads=sample_reads, target_reads=target_reads
    )

    # THEN it passes
    assert passes_reads_threshold


def test_control_total_reads_failing():
    # GIVEN a negative control sample with many reads
    sample_reads = 100
    target_reads = 100

    # WHEN checking if the control read count is valid
    passes_reads_threshold: bool = is_valid_total_reads_for_negative_control(
        reads=sample_reads, target_reads=target_reads
    )

    # THEN it fails
    assert not passes_reads_threshold


def test_control_total_reads_passing_without_reads():
    # GIVEN a negative control sample without reads
    sample_reads = 0
    target_reads = 100

    # WHEN checking if the control read count is valid
    passes_reads_threshold: bool = is_valid_total_reads_for_negative_control(
        reads=sample_reads, target_reads=target_reads
    )

    # THEN it passes
    assert passes_reads_threshold


def test_is_valid_mapping_rate_passing():
    # GIVEN a high mapping rate
    mapping_rate = 0.99

    # WHEN checking if the mapping rate is valid
    passes_mapping_rate_threshold: bool = is_valid_mapping_rate(mapping_rate)

    # THEN it passes
    assert passes_mapping_rate_threshold


def test_is_valid_mapping_rate_failing():
    # GIVEN a low mapping rate
    mapping_rate = 0.1

    # WHEN checking if the mapping rate is valid
    passes_mapping_rate_threshold: bool = is_valid_mapping_rate(mapping_rate)

    # THEN it fails
    assert not passes_mapping_rate_threshold


def test_is_valid_duplication_rate_passing():
    # GIVEN a low duplication rate
    duplication_rate = 0.1

    # WHEN checking if the duplication rate is valid
    passes_duplication_qc: bool = is_valid_duplication_rate(duplication_rate)

    # THEN it passes
    assert passes_duplication_qc


def test_is_valid_duplication_rate_failing():
    # GIVEN a high duplication rate
    duplication_rate = 0.9

    # WHEN checking if the duplication rate is valid
    passes_duplication_qc: bool = is_valid_duplication_rate(duplication_rate)

    # THEN it fails
    assert not passes_duplication_qc


def test_is_valid_median_insert_size_passing():
    # GIVEN a high median insert size
    insert_size = 1000

    # WHEN checking if the median insert size is valid
    passes_insert_size_qc: bool = is_valid_median_insert_size(insert_size)

    # THEN it passes
    assert passes_insert_size_qc


def test_is_valid_median_insert_size_failing():
    # GIVEN a low median insert size
    insert_size = 10

    # WHEN checking if the median insert size is valid
    passes_insert_size_qc = is_valid_median_insert_size(insert_size)

    # THEN it fails
    assert not passes_insert_size_qc


def test_is_valid_average_coverage_passing():
    # GIVEN a high average coverage
    average_coverage = 50

    # WHEN checking if the average coverage is valid
    passes_average_coverage_qc: bool = is_valid_average_coverage(average_coverage)

    # THEN it passes
    assert passes_average_coverage_qc


def test_is_valid_average_coverage_failing():
    # GIVEN a low average coverage
    average_coverage = 1

    # WHEN checking if the average coverage is valid
    passes_average_coverage_qc: bool = is_valid_average_coverage(average_coverage)

    # THEN it fails
    assert not passes_average_coverage_qc


def test_is_valid_10x_coverage_passing():
    # GIVEN a high percent of bases covered at 10x
    coverage_10x = 0.95

    # WHEN checking if the coverage is valid
    passes_coverage_10x_qc: bool = is_valid_10x_coverage(coverage_10x)

    # THEN it passes
    assert passes_coverage_10x_qc


def test_is_valid_10x_coverage_failing():
    # GIVEN a low percent of bases covered at 10x
    coverage_10x = 0.1

    # WHEN checking if the coverage is valid
    passes_coverage_10x_qc: bool = is_valid_10x_coverage(coverage_10x)

    # THEN it fails
    assert not passes_coverage_10x_qc


def test_has_valid_mapping_rate_passing():
    # GIVEN metrics with a high mapping rate
    metrics: SampleMetrics = create_sample_metrics(mapped_rate=0.8)

    # WHEN checking if the mapping rate is valid
    passes_mapping_rate_qc: bool = has_valid_mapping_rate(metrics)

    # THEN it passes the quality control
    assert passes_mapping_rate_qc


def test_has_valid_mapping_rate_missing():
    # GIVEN metrics without a mapping rate
    metrics: SampleMetrics = create_sample_metrics(mapped_rate=None)

    # WHEN checking if the mapping rate is valid
    passes_mapping_rate_qc: bool = has_valid_mapping_rate(metrics)

    # THEN it fails the quality control
    assert not passes_mapping_rate_qc


def test_has_valid_duplication_rate_passing():
    # GIVEN metrics with a low duplication rate
    metrics: SampleMetrics = create_sample_metrics(duplication_rate=0.1)

    # WHEN checking if the duplication rate is valid
    passes_duplication_rate_qc: bool = has_valid_duplication_rate(metrics)

    # THEN it passes the quality control
    assert passes_duplication_rate_qc


def test_has_valid_duplication_rate_missing():
    # GIVEN metrics without a duplication rate
    metrics: SampleMetrics = create_sample_metrics(duplication_rate=None)

    # WHEN checking if the duplication rate is valid
    passes_duplication_rate_qc: bool = has_valid_duplication_rate(metrics)

    # THEN it fails the quality control
    assert not passes_duplication_rate_qc


def test_has_valid_median_insert_size_passing():
    # GIVEN metrics with a high median insert size
    metrics: SampleMetrics = create_sample_metrics(insert_size=200)

    # WHEN checking if the median insert size is valid
    passes_insert_size_qc: bool = has_valid_median_insert_size(metrics)

    # THEN it passes the quality control
    assert passes_insert_size_qc


def test_has_valid_median_insert_size_missing():
    # GIVEN metrics without a median insert size
    metrics: SampleMetrics = create_sample_metrics(insert_size=None)

    # WHEN checking if the median insert size is valid
    passes_insert_size_qc: bool = has_valid_median_insert_size(metrics)

    # THEN it fails the quality control
    assert not passes_insert_size_qc


def test_has_valid_average_coverage_passes():
    # GIVEN metrics with a high average coverage
    metrics: SampleMetrics = create_sample_metrics(average_coverage=30.0)

    # WHEN checking if the average coverage is valid
    passes_average_coverage_qc: bool = has_valid_average_coverage(metrics)

    # THEN it passes the quality control
    assert passes_average_coverage_qc


def test_has_valid_average_coverage_missing():
    # GIVEN metrics without an average coverage
    metrics: SampleMetrics = create_sample_metrics(average_coverage=None)

    # WHEN checking if the average coverage is valid
    passes_average_coverage_qc: bool = has_valid_average_coverage(metrics)

    # THEN it fails the quality control
    assert not passes_average_coverage_qc


def test_has_valid_10x_coverage_passing():
    # GIVEN metrics with a high percent of bases covered at 10x
    metrics: SampleMetrics = create_sample_metrics(coverage_10x=95.0)

    # WHEN checking if the coverage is valid
    passes_coverage_10x_qc: bool = has_valid_10x_coverage(metrics)

    # THEN it passes the quality control
    assert passes_coverage_10x_qc


def test_has_valid_10x_coverage_missing():
    # GIVEN metrics without a percent of bases covered at 10x
    metrics: SampleMetrics = create_sample_metrics(coverage_10x=None)

    # WHEN checking if the coverage is valid
    passes_coverage_10x_qc: bool = has_valid_10x_coverage(metrics)

    # THEN it fails the quality control
    assert not passes_coverage_10x_qc


def test_negative_control_passes_qc():
    # GIVEN a negative control sample that passes quality control
    control_result: SampleQualityResult = create_quality_result(is_control=True)
    other_result: SampleQualityResult = create_quality_result(passes_qc=False)

    # WHEN checking if the negative control passes quality control
    control_passes_qc: bool = negative_control_pass_qc([other_result, control_result])

    # THEN it passes quality control
    assert control_passes_qc


def test_negative_control_fails_qc():
    # GIVEN a negative control sample that fails quality control
    control_result: SampleQualityResult = create_quality_result(is_control=True, passes_qc=False)
    other_result: SampleQualityResult = create_quality_result()

    # WHEN checking if the negative control passes quality control
    control_passes_qc: bool = negative_control_pass_qc([other_result, control_result])

    # THEN it fails quality control
    assert not control_passes_qc


def test_get_urgent_results():
    # GIVEN quality results with urgent and non-urgent samples
    urgent_result: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWRNXTR003, passes_qc=True
    )
    non_urgent_result: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWXNXTR003, passes_qc=True
    )
    quality_results: list[SampleQualityResult] = [urgent_result, non_urgent_result]

    # WHEN getting the urgent results
    urgent_results: list[SampleQualityResult] = get_urgent_results(quality_results)

    # THEN the urgent results are returned
    assert urgent_results == [urgent_result]


def test_urgent_samples_pass_qc():
    # GIVEN quality results with urgent samples that pass quality control
    urgent_result: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWRNXTR003, passes_qc=True
    )
    urgent_result_control: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWRNXTR003, passes_qc=True, is_control=True
    )
    urgent_results: list[SampleQualityResult] = [urgent_result, urgent_result_control]

    # WHEN checking if the urgent samples pass quality control
    urgent_pass_qc: bool = urgent_samples_pass_qc(urgent_results)

    # THEN it passes quality control
    assert urgent_pass_qc


def test_urgent_samples_fail_qc():
    # GIVEN quality results with urgent samples that fail quality control
    urgent_result: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWRNXTR003, passes_qc=False
    )
    urgent_result_control: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWRNXTR003, passes_qc=True, is_control=True
    )
    urgent_results: list[SampleQualityResult] = [urgent_result, urgent_result_control]

    # WHEN checking if the urgent samples pass quality control
    urgent_pass_qc: bool = urgent_samples_pass_qc(urgent_results)

    # THEN it fails quality control
    assert not urgent_pass_qc


def test_get_non_urgent_results():
    # GIVEN quality results with urgent and non-urgent samples
    urgent_result: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWRNXTR003, passes_qc=True
    )
    non_urgent_result: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWXNXTR003, passes_qc=True
    )
    quality_results: list[SampleQualityResult] = [urgent_result, non_urgent_result]

    # WHEN getting the non-urgent results
    non_urgent_results: list[SampleQualityResult] = get_non_urgent_results(quality_results)

    # THEN the non-urgent results are returned
    assert non_urgent_results == [non_urgent_result]


def test_non_urgent_samples_pass_qc():
    # GIVEN quality results with non-urgent samples that pass quality control
    non_urgent_result: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWXNXTR003, passes_qc=True
    )
    non_urgent_result_control: SampleQualityResult = create_quality_result(
        application_tag=MicrosaltAppTags.MWXNXTR003, passes_qc=True, is_control=True
    )
    non_urgent_results: list[SampleQualityResult] = [non_urgent_result, non_urgent_result_control]

    # WHEN checking if the non-urgent samples pass quality control
    non_urgent_pass_qc: bool = non_urgent_samples_pass_qc(non_urgent_results)

    # THEN it passes quality control
    assert non_urgent_pass_qc



"""Tests the findbusinessdata part of the Cg store API related to Analysis model."""

from datetime import datetime

from sqlalchemy.orm import Query

from cg.constants import Workflow
from cg.constants.constants import CaseActions
from cg.constants.subject import PhenotypeStatus
from cg.store.models import Analysis, Case, CaseSample, Sample
from cg.store.store import Store
from tests.store_helpers import StoreHelpers


def test_get_latest_nipt_analysis_to_upload(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store,
    timestamp_now: datetime,
    workflow: str = Workflow.FLUFFY,
):
    """Test get the latest NIPT analysis to upload."""
    # GIVEN an analysis that is not delivery reported but there exists a newer analysis

    # WHEN fetching the latest analysis to upload to nipt
    analyses: list[Analysis] = (
        store_with_analyses_for_cases_not_uploaded_fluffy.get_latest_analysis_to_upload_for_workflow(
            workflow=workflow
        )
    )

    # THEN only the newest analysis should be returned
    for analysis in analyses:
        assert analysis.started_at == timestamp_now
        assert analysis.uploaded_at is None
        assert analysis.pipeline == workflow


def test_get_latest_microsalt_analysis_to_upload(
    store_with_analyses_for_cases_not_uploaded_microsalt: Store,
    timestamp_now: datetime,
    workflow: str = Workflow.MICROSALT,
):
    """Test get the latest microsalt analysis to upload."""
    # GIVEN an analysis that is not delivery reported but there exists a newer analysis

    # WHEN fetching the latest analysis to upload to microsalt
    analyses: list[Analysis] = (
        store_with_analyses_for_cases_not_uploaded_microsalt.get_latest_analysis_to_upload_for_workflow(
            workflow=workflow
        )
    )

    # THEN only the newest analysis should be returned
    for analysis in analyses:
        assert analysis.started_at == timestamp_now
        assert analysis.uploaded_at is None
        assert analysis.pipeline == workflow


def test_get_analyses_to_deliver_for_pipeline(
    store_with_analyses_for_cases_to_deliver: Store,
    workflow: Workflow = Workflow.FLUFFY,
):
    # GIVEN a store with multiple analyses to deliver

    # WHEN fetching the latest analysis to upload to nipt
    analyses = store_with_analyses_for_cases_to_deliver.get_analyses_to_deliver_for_pipeline(
        workflow=workflow
    )

    # THEN only the newest analysis should be returned
    for analysis in analyses:
        assert analysis.case.internal_id in ["test_case_1", "yellowhog"]
        assert analysis.uploaded_at is None
        assert analysis.pipeline == workflow


def test_get_analyses(store_with_analyses_for_cases: Store):
    """Test all analyses can be returned."""
    # GIVEN a database with an analysis and case

    # WHEN fetching all analyses
    analysis: list[Analysis] = store_with_analyses_for_cases.get_analyses()

    # THEN one analysis should be returned
    assert len(analysis) == store_with_analyses_for_cases._get_query(table=Analysis).count()


def test_get_families_with_extended_models(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a query is returned from the database."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(
        base_store, completed_at=timestamp_now, workflow=Workflow.MIP_DNA
    )

    # Given an action set to analyze
    test_analysis.case.action: str = CaseActions.ANALYZE

    # GIVEN a database with a case with one of sequenced samples and completed analysis
    link = base_store.relate_sample(test_analysis.case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting cases to analyse
    cases: list[Query] = list(base_store._get_outer_join_cases_with_analyses_query())

    case: Case = cases[0]

    # THEN cases should be returned
    assert cases

    # THEN analysis should be part of cases attributes
    assert case.analyses[0].pipeline == Workflow.MIP_DNA


def test_get_families_with_extended_models_when_no_case(base_store: Store):
    """test that no case is returned from the database when no cases."""

    # GIVEN an empty database

    # WHEN getting cases to analyse
    cases: list[Query] = list(base_store._get_outer_join_cases_with_analyses_query())

    # THEN no cases should be returned
    assert not cases


def test_get_cases_with_samples_query(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case and samples query is returned from the database."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(
        base_store, completed_at=timestamp_now, workflow=Workflow.MIP_DNA
    )

    # GIVEN a database with a case with one of sequenced samples and completed analysis
    link = base_store.relate_sample(test_analysis.case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting the stored case with its associated samples
    cases: list[Query] = list(base_store._get_join_cases_with_samples_query())

    # THEN a list of cases should be returned, and it should contain the stored and linked sample
    assert cases
    assert test_sample == cases[0].links[0].sample


def test_that_many_cases_can_have_one_sample_each(
    base_store: Store, helpers: StoreHelpers, max_nr_of_cases: int, timestamp_now: datetime
):
    """Test that tests that cases are returned even if there are many result rows in the query."""

    # GIVEN a database with max_nr_of_cases cases
    test_cases: list[Case] = helpers.add_cases_with_samples(
        base_store, max_nr_of_cases, sequenced_at=timestamp_now
    )

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.MIP_DNA)

    # THEN cases should contain all cases since they are to be analysed
    assert len(cases) == len(test_cases)


def test_that_cases_can_have_many_samples(
    base_store: Store, helpers, max_nr_of_samples: int, timestamp_now: datetime
):
    """Test that tests that cases are returned even if there are many result rows in the query."""

    # GIVEN a cases with max_nr_of_samples sequenced samples
    case_with_50: Case = helpers.add_case_with_samples(
        base_store, "case_with_50_samples", max_nr_of_samples, sequenced_at=timestamp_now
    )

    # GIVEN a sequnced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)
    assert test_sample.last_sequenced_at

    # GIVEN a case with one sample
    case_with_one: Case = helpers.add_case(base_store, "case_with_one_sample")

    # GIVEN a database with a case with one sample sequenced sample
    link = base_store.relate_sample(case_with_one, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.MIP_DNA)

    # THEN cases should be returned
    assert cases

    # THEN cases should contain the test case
    assert case_with_50 in cases
    assert case_with_one in cases


def test_external_sample_to_re_analyse(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case marked for re-analysis with one sample external not sequenced in-house and
    with completed analysis show up among the cases to analyse."""

    # GIVEN a sample which is not sequenced and external
    test_sample: Sample = helpers.add_sample(base_store, is_external=True, last_sequenced_at=None)

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(
        base_store, completed_at=timestamp_now, workflow=Workflow.MIP_DNA
    )
    assert test_analysis.completed_at

    # Given an action set to analyze
    test_analysis.case.action: str = CaseActions.ANALYZE

    # GIVEN a database with a case with one not sequenced external sample
    link = base_store.relate_sample(test_analysis.case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.MIP_DNA)

    # THEN cases should be returned
    assert cases

    # THEN test case should be among the cases returned for analysis
    assert test_analysis.case in cases


def test_new_external_case_not_in_result(base_store: Store, helpers: StoreHelpers):
    """Test that a case with one external sample that has no specified data_analysis does not show up."""

    # GIVEN an externally sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, is_external=True, last_sequenced_at=None)

    # GIVEN a cancer case
    test_case: Case = helpers.add_case(base_store, data_analysis=Workflow.BALSAMIC)

    # GIVEN a database with a case with one externally sequenced samples for BALSAMIC analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.BALSAMIC)

    # THEN cases should not contain the test case
    assert test_case not in cases


def test_case_to_re_analyse(base_store: Store, helpers: StoreHelpers, timestamp_now: datetime):
    """Test that a case marked for re-analyse with one sample that has been sequenced and
    with completed analysis do show up among the cases to analyse."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(
        base_store, completed_at=timestamp_now, workflow=Workflow.MIP_DNA
    )

    # Given an action set to analyze
    test_analysis.case.action: str = CaseActions.ANALYZE

    # GIVEN a database with a case with one of one sequenced samples and completed analysis
    link = base_store.relate_sample(test_analysis.case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.MIP_DNA)

    # THEN cases should be returned
    assert cases

    # THEN test case should be among the cases returned for analysis
    assert test_analysis.case in cases


def test_all_samples_and_analysis_completed(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case with one sample that has been sequenced and with completed
    analysis don't show up among the cases to analyse."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a completed analysis
    test_analysis: Analysis = helpers.add_analysis(base_store, completed_at=timestamp_now)

    # Given a completed analysis
    test_analysis.case.action: str | None = None

    # GIVEN a database with a case with one of one sequenced samples and completed analysis
    link = base_store.relate_sample(test_analysis.case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.MIP_DNA)

    # THEN cases should not contain the test case
    assert not cases


def test_specified_analysis_in_result(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case with one sample that has specified data_analysis does show up."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a cancer case
    test_case: Case = helpers.add_case(base_store, data_analysis=Workflow.BALSAMIC)

    # GIVEN a database with a case with one sequenced samples for BALSAMIC analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.BALSAMIC)

    # THEN cases should be returned
    assert cases

    # THEN cases should contain the test case
    assert test_case in cases


def test_exclude_other_pipeline_analysis_from_result(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case with specified analysis and with one sample does not show up among
    others."""

    # GIVEN a sequenced sample
    test_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a cancer case
    test_case = helpers.add_case(base_store, data_analysis=Workflow.BALSAMIC)

    # GIVEN a database with a case with one sequenced samples for specified analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)

    # WHEN getting cases to analyse for another workflow
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.MIP_DNA)

    # THEN cases should not contain the test case
    assert test_case not in cases


def test_one_of_two_sequenced_samples(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case with one sequenced sample and one not sequenced sample do not show up among the
    cases to analyse."""

    # GIVEN a case
    test_case: Case = helpers.add_case(base_store)

    # GIVEN a sequenced sample
    sequenced_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a NOT sequenced sample
    not_sequenced_sample: Sample = helpers.add_sample(base_store, last_sequenced_at=None)

    # GIVEN a database with a case with one of one sequenced samples and no analysis
    link_1: CaseSample = base_store.relate_sample(
        test_case, sequenced_sample, PhenotypeStatus.UNKNOWN
    )
    link_2: CaseSample = base_store.relate_sample(
        test_case, not_sequenced_sample, PhenotypeStatus.UNKNOWN
    )
    base_store.session.add_all([link_1, link_2])

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.MIP_DNA, threshold=True)

    # THEN no cases should be returned
    assert not cases


def test_one_of_one_sequenced_samples(
    base_store: Store, helpers: StoreHelpers, timestamp_now: datetime
):
    """Test that a case with one of one samples that has been sequenced shows up among the
    cases to analyse."""

    # GIVEN a case
    test_case: Case = helpers.add_case(base_store)

    # GIVEN a sequenced sample
    test_sample = helpers.add_sample(base_store, last_sequenced_at=timestamp_now)

    # GIVEN a database with a case with a sequenced samples and no analysis
    link = base_store.relate_sample(test_case, test_sample, PhenotypeStatus.UNKNOWN)
    base_store.session.add(link)
    assert test_sample.last_sequenced_at is not None

    # WHEN getting cases to analyse
    cases: list[Case] = base_store.cases_to_analyze(workflow=Workflow.MIP_DNA)

    # THEN cases should be returned
    assert cases

    # THEN cases should contain the test case
    assert test_case in cases


def test_get_analyses_for_case_and_pipeline_before(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store,
    timestamp_now: datetime,
    workflow: Workflow = Workflow.FLUFFY,
    case_id: str = "yellowhog",
):
    """Test to get all analyses before a given date."""

    # GIVEN a database with a number of analyses

    # WHEN getting all analyses before a given date
    analyses: list[Analysis] = (
        store_with_analyses_for_cases_not_uploaded_fluffy.get_analyses_for_case_and_workflow_started_at_before(
            workflow=workflow, started_at_before=timestamp_now, case_internal_id=case_id
        )
    )

    # THEN assert that the analyses before the given date are returned
    for analysis in analyses:
        assert analysis.started_at < timestamp_now
        assert analysis.case.internal_id == case_id
        assert analysis.pipeline == workflow


def test_get_analyses_for_case_before(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store,
    timestamp_now: datetime,
    case_id: str = "yellowhog",
):
    """Test to get all analyses before a given date."""

    # GIVEN a database with a number of analyses

    # WHEN getting all analyses before a given date
    analyses: list[Analysis] = (
        store_with_analyses_for_cases_not_uploaded_fluffy.get_analyses_for_case_started_at_before(
            case_internal_id=case_id,
            started_at_before=timestamp_now,
        )
    )

    # THEN assert that the analyses before the given date are returned
    for analysis in analyses:
        assert analysis.started_at < timestamp_now
        assert analysis.case.internal_id == case_id


def test_get_analyses_for_pipeline_before(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store,
    timestamp_now: datetime,
    workflow: Workflow = Workflow.FLUFFY,
):
    """Test to get all analyses for a workflow before a given date."""

    # GIVEN a database with a number of analyses

    # WHEN getting all analyses before a given date
    analyses: list[Analysis] = (
        store_with_analyses_for_cases_not_uploaded_fluffy.get_analyses_for_workflow_started_at_before(
            workflow=workflow, started_at_before=timestamp_now
        )
    )

    # THEN assert that the analyses before the given date are returned
    for analysis in analyses:
        assert analysis.started_at < timestamp_now
        assert analysis.pipeline == workflow


def test_get_analyses_before(
    store_with_analyses_for_cases_not_uploaded_fluffy: Store,
    timestamp_now: datetime,
):
    """Test to get all analyses for a workflow before a given date."""

    # GIVEN a database with a number of analyses

    # WHEN getting all analyses before a given date
    analyses: list[Analysis] = (
        store_with_analyses_for_cases_not_uploaded_fluffy.get_analyses_started_at_before(
            started_at_before=timestamp_now
        )
    )

    # THEN assert that the analyses before the given date are returned
    for analysis in analyses:
        assert analysis.started_at < timestamp_now



from enum import StrEnum


class AnalysisStatus:
    CANCELLED: str = "cancelled"
    COMPLETED: str = "completed"
    ERROR: str = "error"
    FAILED: str = "failed"
    PENDING: str = "pending"
    RUNNING: str = "running"
    TIMEOUT: str = "timeout"
    QC: str = "qc"


class AnalysisTypes(StrEnum):
    WGS: str = "wgs"
    WES: str = "wes"
    TGS: str = "tgs"
    RNA: str = "rna"
    WTS: str = "wts"
    OTHER: str = "other"



import gzip
from pathlib import Path


def read_gzip_first_line(file_path: Path) -> str:
    """Return first line of gzip file."""
    with gzip.open(file_path) as file:
        return file.readline().decode()



from abc import ABC, abstractmethod

from cg.models.slurm.sbatch import Sbatch


class SlurmService(ABC):
    @abstractmethod
    def submit_job(self, job_config: Sbatch) -> int:
        """Submit a job and return the job id."""
        pass



"""Gene panel specific constants."""

from enum import StrEnum

from cg.constants.constants import CustomerId

GENOME_BUILD_37: str = "37"
GENOME_BUILD_38: str = "GRCh38"


class GenePanelMasterList(StrEnum):
    BRAIN: str = "BRAIN"
    CARDIOLOGY: str = "Cardiology"
    CILM: str = "CILM"
    CH: str = "CH"
    CTD: str = "CTD"
    DIAB: str = "DIAB"
    ENDO: str = "ENDO"
    EP: str = "EP"
    HEARING: str = "HEARING"
    HYDRO: str = "HYDRO"
    IBMFS: str = "IBMFS"
    IEM: str = "IEM"
    IF: str = "IF"
    MCARTA: str = "mcarta"
    MHT: str = "MHT"
    MIT: str = "MIT"
    MOVE: str = "MOVE"
    MTDNA: str = "mtDNA"
    NBS_M: str = "NBS-M"
    NEURODEG: str = "NEURODEG"
    NMD: str = "NMD"
    OMIM_AUTO: str = "OMIM-AUTO"
    OPTIC: str = "OPTIC"
    PANELAPP_GREEN: str = "PANELAPP-GREEN"
    PEDHEP: str = "PEDHEP"
    PID: str = "PID"
    PIDCAD: str = "PIDCAD"
    RETINA: str = "RETINA"
    SKD: str = "SKD"
    SOVM: str = "SOVM"
    STROKE: str = "STROKE"
    AID: str = "AID"
    INHERITED_CANCER: str = "Inherited cancer"

    @classmethod
    def get_panel_names(cls, panels=None) -> list[str]:
        """Return requested panel names from the Master list, or all panels if none are specified."""
        return list(panels) if panels else list(map(lambda panel: panel.value, cls))

    @staticmethod
    def collaborators() -> set[str]:
        """Return collaborators of the Master list."""
        return {
            CustomerId.CG_INTERNAL_CUSTOMER,
            CustomerId.CUST002,
            CustomerId.CUST003,
            CustomerId.CUST004,
            CustomerId.CUST042,
        }


class GenePanelCombo:
    COMBO_1: dict[str, set[str]] = {
        "DSD": {"DSD", "DSD-S", "HYP", "SEXDIF", "SEXDET"},
        "CM": {"CNM", "CM"},
        "Horsel": {"Horsel", "141217", "141201"},
        "OPHTHALMO": {
            "OPHTHALMO",
            "ANTE-ED",
            "CATARACT",
            "CORNEA",
            "GLAUCOMA",
            "RETINA",
            "SED",
            "ALBINISM",
        },
    }



"""Module to handle NF executors."""

from datetime import datetime
from pathlib import Path
from typing import Iterable

from cg.apps.slurm.slurm_api import SlurmAPI
from cg.constants.constants import FileExtensions, FileFormat
from cg.constants.nextflow import (
    JAVA_MEMORY_HEADJOB,
    NXF_JVM_ARGS_ENV,
    SlurmHeadJobDefaults,
)
from cg.io.controller import ReadFile
from cg.models.slurm.sbatch import Sbatch
from cg.utils.utils import build_command_from_dict


class NfBaseHandler:
    """
    Parent class for handling the interaction with NF executors that are common to both Nextflow and NF-Tower.
    """

    pass


class NfTowerHandler(NfBaseHandler):
    """
    Parent class for handling the interaction with NF-Tower.
    """

    @classmethod
    def get_tower_launch_parameters(cls, tower_workflow: str, command_args: dict) -> list[str]:
        """Returns a tower launch command given a dictionary with arguments."""

        tower_options: list[str] = build_command_from_dict(
            options={
                f"--{arg.replace('_', '-')}": command_args.get(arg, None)
                for arg in (
                    "work_dir",
                    "profile",
                    "params_file",
                    "config",
                    "name",
                    "revision",
                    "compute_env",
                )
            },
            exclude_true=True,
        )
        return ["launch"] + tower_options + [tower_workflow]

    @classmethod
    def get_tower_relaunch_parameters(cls, from_tower_id: int, command_args: dict) -> list[str]:
        """Returns a tower relaunch command given a dictionary with arguments."""

        tower_options: list[str] = build_command_from_dict(
            options={
                f"--{arg.replace('_', '-')}": command_args.get(arg, None)
                for arg in (
                    "profile",
                    "params_file",
                    "config",
                    "compute_env",
                )
            },
            exclude_true=True,
        )
        now: str = datetime.now().strftime("%Y%m%d%H%M%S")
        return [
            "runs",
            "relaunch",
            "--id",
            str(from_tower_id),
            "--name",
            f"{command_args.get('name')}_from{str(from_tower_id)}_{now}",
        ] + tower_options

    @staticmethod
    def get_tower_id(stdout_lines: Iterable) -> str:
        """Parse the stdout and return a workflow id. An example of the output to parse is:
        Case <CASE_ID> exists in status db
        Running RNAFUSION analysis for <CASE_ID>
        Workflow will be executed using tower
        Running command <COMMAND>

          Workflow 1uxZE9JM7Tl58r submitted at [<WORKSPACE>] workspace.

        https://<URL_TO_TOWER_CASE>
        Action running set for case <CASE_ID>"""
        for line in stdout_lines:
            if line.strip().startswith("Workflow"):
                return line.strip().split()[1]

    @classmethod
    def get_last_tower_id(cls, case_id: str, trailblazer_config: Path) -> int:
        """Return the previously-stored NF-Tower ID."""
        if not trailblazer_config.exists():
            raise FileNotFoundError(f"No NF-Tower ID found for case {case_id}.")
        return ReadFile.get_content_from_file(
            file_format=FileFormat.YAML, file_path=trailblazer_config
        ).get(case_id)[-1]


class NextflowHandler(NfBaseHandler):
    """
    Parent class for handling the interaction with Nextflow.
    """

    @staticmethod
    def get_variables_to_export() -> dict[str, str]:
        """Dictionary with required environment variables to be exported."""
        return {NXF_JVM_ARGS_ENV: f"'{JAVA_MEMORY_HEADJOB}'"}

    @classmethod
    def get_nextflow_run_parameters(
        cls, case_id: str, workflow_path: str, root_dir: str, command_args: dict
    ) -> list[str]:
        """Returns a Nextflow run command given a dictionary with arguments."""

        nextflow_options: list[str] = build_command_from_dict(
            options=dict((f"-{arg}", command_args.get(arg, True)) for arg in ("log", "config")),
            exclude_true=True,
        )
        run_options: list[str] = build_command_from_dict(
            options=dict(
                (f"-{arg.replace('_', '-')}", command_args.get(arg, None))
                for arg in (
                    "work_dir",
                    "resume",
                    "profile",
                    "with_tower",
                    "params_file",
                )
            ),
            exclude_true=True,
        )
        return nextflow_options + ["run", workflow_path] + run_options

    @staticmethod
    def get_head_job_sbatch_path(case_directory: Path) -> Path:
        """Path to Nextflow sbatch for the head job."""
        return Path(case_directory, "nextflow_head_job").with_suffix(FileExtensions.SBATCH)

    @classmethod
    def execute_head_job(
        cls,
        case_id: str,
        case_directory: Path,
        slurm_account: str,
        email: str,
        qos: str,
        commands: str,
        hours: int = SlurmHeadJobDefaults.HOURS,
        memory: int = SlurmHeadJobDefaults.MEMORY,
        number_tasks: int = SlurmHeadJobDefaults.NUMBER_TASKS,
        dry_run: bool = False,
    ) -> int:
        """Executes Nextflow head job command."""

        slurm_api: SlurmAPI = SlurmAPI()
        slurm_api.set_dry_run(dry_run=dry_run)
        sbatch_parameters: Sbatch = Sbatch(
            account=slurm_account,
            commands=commands,
            email=email,
            hours=hours,
            job_name=f"{case_id}.%j",
            log_dir=case_directory.as_posix(),
            memory=memory,
            number_tasks=number_tasks,
            quality_of_service=qos,
        )

        sbatch_content: str = slurm_api.generate_sbatch_content(sbatch_parameters=sbatch_parameters)
        sbatch_path: Path = cls.get_head_job_sbatch_path(case_directory=case_directory)
        sbatch_number: int = slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        return sbatch_number



"""Tests CLI common methods to create the case config for NF analyses."""

import logging

import pytest
from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.rnafusion.base import config_case
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "context",
    ["rnafusion_context", "taxprofiler_context"],
)
def test_config_case_without_options(cli_runner: CliRunner, context: CGConfig, request):
    """Test config_case for Taxprofiler and Rnafusion without options."""
    context = request.getfixturevalue(context)

    # WHEN dry running without anything specified
    result = cli_runner.invoke(config_case, obj=context)

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN command log should inform about missing arguments
    assert "Missing argument" in result.output


@pytest.mark.parametrize(
    "context",
    ["rnafusion_context", "taxprofiler_context"],
)
def test_config_case_with_missing_case(
    cli_runner: CliRunner,
    caplog: LogCaptureFixture,
    case_id_does_not_exist: str,
    context: CGConfig,
    request,
):
    """Test config_case for Taxprofiler and Rnafusion with a missing case."""
    caplog.set_level(logging.ERROR)
    context = request.getfixturevalue(context)

    # GIVEN a case not in the StatusDB database
    assert not context.status_db.get_case_by_internal_id(internal_id=case_id_does_not_exist)

    # WHEN running
    result = cli_runner.invoke(config_case, [case_id_does_not_exist], obj=context)

    # THEN command should not exit successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN the error log should indicate that the case is invalid
    assert "could not be found in Status DB!" in caplog.text



"""Test fixtures for cg/server tests"""

import os
from datetime import datetime
from typing import Generator

import pytest
from flask import Flask
from flask.testing import FlaskClient
from mock import patch

from cg.constants import DataDelivery, Workflow
from cg.server.ext import db as store
from cg.store.database import create_all_tables, drop_all_tables
from cg.store.models import Case, Customer, Order
from tests.store_helpers import StoreHelpers

os.environ["CG_SQL_DATABASE_URI"] = "sqlite:///"
os.environ["LIMS_HOST"] = "dummy_value"
os.environ["LIMS_USERNAME"] = "dummy_value"
os.environ["LIMS_PASSWORD"] = "dummy_value"
os.environ["GOOGLE_OAUTH_CLIENT_ID"] = "dummy_value"
os.environ["GOOGLE_OAUTH_CLIENT_SECRET"] = "dummy_value"
os.environ["CG_ENABLE_ADMIN"] = "1"


@pytest.fixture
def app() -> Generator[Flask, None, None]:
    from cg.server.auto import app

    app.config.update({"TESTING": True})
    create_all_tables()
    yield app
    drop_all_tables()


@pytest.fixture
def case(helpers: StoreHelpers) -> Case:
    case: Case = helpers.add_case(
        customer_id=1,
        data_analysis=Workflow.MIP_DNA,
        data_delivery=DataDelivery.ANALYSIS_SCOUT,
        name="test case",
        ticket="123",
        store=store,
    )
    return case


@pytest.fixture
def customer(helpers: StoreHelpers) -> Customer:
    customer: Customer = helpers.ensure_customer(store=store, customer_id="test_customer")
    return customer


@pytest.fixture
def customer_another(helpers: StoreHelpers) -> Customer:
    customer: Customer = helpers.ensure_customer(store=store, customer_id="test_customer_2")
    return customer


@pytest.fixture
def order(helpers: StoreHelpers, customer: Customer) -> Order:
    order: Order = helpers.add_order(
        store=store,
        customer_id=customer.id,
        ticket_id=1,
        order_date=datetime.now(),
    )
    return order


@pytest.fixture
def order_another(helpers: StoreHelpers, customer_another: Customer) -> Order:
    order: Order = helpers.add_order(
        store=store, customer_id=customer_another.id, ticket_id=2, order_date=datetime.now()
    )
    return order


@pytest.fixture
def order_balsamic(helpers: StoreHelpers, customer_another: Customer) -> Order:
    order: Order = helpers.add_order(
        store=store,
        customer_id=customer_another.id,
        ticket_id=3,
        order_date=datetime.now(),
        workflow=Workflow.BALSAMIC,
    )
    return order


@pytest.fixture
def non_existent_order_id() -> int:
    return 900


@pytest.fixture
def client(app: Flask) -> Generator[FlaskClient, None, None]:
    # Bypass authentication
    with patch.object(app, "before_request_funcs", new={}):
        yield app.test_client()



"""CLI options for Nextflow and NF-Tower."""

import click

from cg.cli.workflow.commands import ARGUMENT_CASE_ID, OPTION_DRY
from cg.constants.constants import MetaApis
from cg.exc import CgError
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig

OPTION_WORKDIR = click.option(
    "--work-dir",
    type=click.Path(),
    help="Directory where intermediate result files are stored",
)
OPTION_RESUME = click.option(
    "--resume",
    is_flag=True,
    default=False,
    show_default=True,
    help="Execute the script using the cached results, useful to continue \
        executions that was stopped by an error",
)
OPTION_PROFILE = click.option(
    "--profile",
    type=str,
    show_default=True,
    help="Choose a configuration profile",
)

OPTION_LOG = click.option(
    "--log",
    type=click.Path(),
    help="Set nextflow log file path",
)

OPTION_CONFIG = click.option(
    "--config",
    type=click.Path(),
    help="Nextflow config file path",
)

OPTION_PARAMS_FILE = click.option(
    "--params-file",
    type=click.Path(),
    help="Nextflow workflow-specific parameter file path",
)

OPTION_USE_NEXTFLOW = click.option(
    "--use-nextflow",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help="Execute workflow using nextflow",
)

OPTION_REVISION = click.option(
    "--revision",
    type=str,
    help="Revision of workflow to run (either a git branch, tag or commit SHA number)",
)
OPTION_COMPUTE_ENV = click.option(
    "--compute-env",
    type=str,
    help="Compute environment name. If not specified the primary compute environment will be used.",
)
OPTION_TOWER_RUN_ID = click.option(
    "--nf-tower-id",
    type=str,
    is_flag=False,
    default=None,
    help="NF-Tower ID of run to relaunch. If not provided the latest NF-Tower ID for a case will be used.",
)


@click.command("metrics-deliver")
@ARGUMENT_CASE_ID
@OPTION_DRY
@click.pass_obj
def metrics_deliver(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Create and validate a metrics deliverables file for given case id.
    If QC metrics are met it sets the status in Trailblazer to complete.
    If failed, it sets it as failed and adds a comment with information of the failed metrics."""

    analysis_api: NfAnalysisAPI = context.meta_apis[MetaApis.ANALYSIS_API]

    try:
        analysis_api.status_db.verify_case_exists(case_internal_id=case_id)
        analysis_api.write_metrics_deliverables(case_id=case_id, dry_run=dry_run)
        analysis_api.validate_qc_metrics(case_id=case_id, dry_run=dry_run)
    except CgError as error:
        raise click.Abort() from error
